  #include "nrutil.h"

#include <sys/types.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <float.h>

#include <memory.h>

typedef struct Image
{
  int width;
  int height;
  double *fdata;
} Image;


static void
die(char *message)
{
  fprintf(stderr, "ppm: %s\n", message);
  exit(1);
}


void mglin(double **u, int n, int ncycle);
double **dmatrix(long nrl, long nrh, long ncl, long nch);


void
ImageSetPixelf(Image *image, int x, int y, double val)
{
  int offset = (y * image->width + x);

  image->fdata[offset] = val;
}


double ImageGetPixelf(Image *image, int x, int y)
{
  int offset = (y * image->width + x);

  return image->fdata[offset];
}

void ImageWriteP7grayfloat2color(Image *image, char *filename)
{
  int i, j, count = 0, num;
  int size = image->width * image->height;

  FILE *fp = fopen(filename, "wb");

  if (!fp) die("cannot open file for writing");

  float *fdata   = (float *) malloc(sizeof (float) * 3 * image->width * image->height); // float just for one channel

  fprintf(fp, "PF\n%d %d\n%f\n",  image->width , image->height , -1.0);

  for (j=0; j < image->height; j++) {
	  for (i=0; i < image->width; i++) {
		  fdata[count++] = (float) (ImageGetPixelf(image,i,j));
		  fdata[count++] = (float) (ImageGetPixelf(image,i,j));
		  fdata[count++] = (float) (ImageGetPixelf(image,i,j));
	  }
  }
  printf("Written %s , with count = %d (%d %d)\n", filename, count, image->width, image->height);
  num = fwrite((float *) fdata, sizeof(float), (size_t) count, fp);

  fclose(fp);
}  

Image *ImageReadP7color2gray(char *filename)
{//just takes the average
  int i, j, c, width, height, num, size, r;

  Image *image = (Image *) malloc(sizeof(Image));
  FILE  *fp    = fopen(filename, "rb");

	long pos=ftell(fp);
	fseek(fp,0,SEEK_END);
	pos=ftell(fp);
	fseek(fp,0,SEEK_SET);
	pos=ftell(fp);

  if (!image) die("cannot allocate memory for new image");
  if (!fp)    die("cannot open file for reading");

  char ch;
  float  maxval;

  if (fscanf(fp, "PF%c\n", &ch) != 1)
    die("file is not in ppm raw format; cannot read");

  ch = getc(fp);
  while (ch == '#')
    {
      do {
	ch = getc(fp);
      } while (ch != '\n');
      ch = getc(fp);
    }

  if (!isdigit(ch)) die("cannot read header information from ppm file");

  ungetc(ch, fp);

	char buffer[256];
	memset(buffer,0,256);
	int count=0, pos2=0;
	do
	{
		ch=getc(fp);
		buffer[pos2++]=ch;
		if(ch==0x0A)count++;
		
	}
	while (count<2);

  sscanf(buffer, "%d%d%f\n", &width, &height, &maxval);

  if (maxval != -1.0) die("image is not true-color (24 bit); read failed");

  size          = width * height;
  image->fdata   = (double *) malloc(size * sizeof(double));
  image->width  = width;
  image->height = height;

  if (!image->fdata) die("cannot allocate memory for new image");

  ch = '\0';
//  ungetc(ch, fp);

  float data, sum;

  num=0;

	for (j=0; j < image->height; j++)
		for (i=0; i < image->width; i++)
		{
			sum = 0;
			for(c=0; c<3; c++)
			{
				num++;
				r=fread(&data,sizeof(data),1,fp);
				sum+=data;
			}
			sum /= 3.0;
			if (_isnan(sum))sum=0.0;
			ImageSetPixelf(image,i,j,sum);
		}

  if (num != 3*size) die("cannot read image data from file");

  fclose(fp);

  return image;
}  

Image *
ImageCreate(int width, int height)
{
  Image *image = (Image *) malloc(sizeof(Image));

  if (!image) die("cannot allocate memory for new image");

  image->width  = width;
  image->height = height;
  image->fdata   = (double *) malloc(sizeof (double) * width * height); // double just for one channel

  if (!image->fdata) die("cannot allocate memory for new image");
  for (int i=0; i<width*height; i++)
  {
	  image->fdata[i]=0.0f;
  }

  return image;
}

void ImageDestroy(Image *image)
{
	if(image)
	{
		if(image->fdata) free (image->fdata);
		image->fdata=NULL;
		free (image);
		image=NULL;
	}
}

Image *ImagePadf(Image *original, int width, int height)
{
	double val;
	int xm=(width-original->width)/2;
	int ym=(height-original->height)/2;
	Image *result=ImageCreate(width, height);
	memset(result->fdata,0,width*height*sizeof(double));
	for(int x=0; x<original->width; x++)
		for(int y=0; y<original->height; y++)
			{
				val=ImageGetPixelf(original,x,y);
				ImageSetPixelf(result,x+xm,y+ym,val);
			}
	return result;
}

Image *ImageCropf(Image *original, int width, int height)
{
	double val;
	int xm=-(width-original->width)/2;
	int ym=-(height-original->height)/2;
	Image *result=ImageCreate(width, height);
	memset(result->fdata,0,width*height*sizeof(double));
	for(int x=0; x<width; x++)
		for(int y=0; y<height; y++)
			{
				val=ImageGetPixelf(original,x+xm,y+ym);
				ImageSetPixelf(result,x,y,val);
			}
	return result;
}

//----------------------------------------------------------------------------------------
Image *ImageIntegrate(Image *gx, Image *gy )
{
	int Vcylces = 5;

	int width = gx->width, height = gx->height;
	double **u=dmatrix(0,width,0,height);
	if (!u) nrerror("allocation failure 1 in matrix()");
	int i,j;

	Image *divG = ImageCreate(width,height);
	double gradSumX, gradSumY;

	for (i=0; i < width; i++) {
		for (j=0; j < height; j++) {
			
			if (i == 0) gradSumX = 0;
			else gradSumX = (double) (ImageGetPixelf(gx, i, j) - ImageGetPixelf(gx, i-1, j));
			if (j == 0) gradSumY = 0;
			else gradSumY = (double) (ImageGetPixelf(gy, i, j) - ImageGetPixelf(gy, i, j-1));
			
			ImageSetPixelf(divG, i, j, (gradSumX + gradSumY)) ; 
		}
	}

	for (i=0; i < width; i++) {
		for (j=0; j < height; j++) {
			u[i+1][j+1] = (double) (ImageGetPixelf(divG, i, j));
		}
	}
	ImageDestroy(divG);

	mglin((double **) u, gx->width-1, Vcylces);	printf("\tDone with mgline fn\n\n"); // Full Multigrid Algorithm using linear method

	Image *out = ImageCreate(width,height);
	for (i=1; i < width; i++) {//not <= to ignore edges
		for (j=1; j < height; j++) {
			ImageSetPixelf(out, i-1, j-1, u[i][j]) ;
		}
	}
	free_dmatrix(u,0,gx->width,0,gy->height);
	return out;
}

//find the closest power of 2
int closestP2(int n1, int n2)
{
	int r=1;
	while ((r<n1)||(r<n2))
	{
		r*=2;
	}
	return r+2;
}

int main(int argc, char *argv[])
{
	Image* dx=ImageReadP7color2gray("1.pfm");
	int size=closestP2(dx->width,dx->height);
	Image*dx2 = ImagePadf(dx,size,size);
//	ImageWriteP7grayfloat2color(dx2,"1p.pfm");
	Image* dy=ImageReadP7color2gray("2.pfm");
	Image*dy2 = ImagePadf(dy,size,size);
//	ImageWriteP7grayfloat2color(dy2,"2p.pfm");
	Image* res=ImageIntegrate(dx2,dy2);
	Image* res2=ImageCropf(res,dx->width,dx->height);
//	ImageWriteP7grayfloat2color(res,"r0.pfm");
	ImageWriteP7grayfloat2color(res2,"r.pfm");
}
